#
# The MIT License
#
# @copyright Copyright (c) 2017 Intel Corporation
# @copyright Copyright (c) 2021 ApertureData Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
from __future__ import annotations
import os
import requests
import time
import json
import logging

from types import SimpleNamespace
from aperturedb.Connector import Connector
from aperturedb.Configuration import Configuration

logger = logging.getLogger(__name__)

PROTOCOL_VERSION = 1


class ConnectorRest(Connector):
    """
    **Class to use ApertureDB's REST interface**

    This class is used to connect to ApertureDB using the REST interface.

    Args:
        str (host): Address of the host to connect to.
        int (port): Port to connect to.
        str (user): Username to specify while establishing a connection.
        str (password): Password to specify while connecting to ApertureDB.
        str (token): Token to use while connecting to the database.
        bool (use_ssl): Use SSL to encrypt communication with the database.
    """

    def __init__(self, host="localhost", port=None,
                 user="", password="", token="",
                 use_ssl=True, shared_data=None,
                 config: Configuration = None):
        self.use_keepalive = False
        super().__init__(
            host=host,
            port=port,
            user=user,
            password=password,
            token=token,
            use_ssl=use_ssl,
            shared_data=shared_data,
            config=config)

        if port is None:
            self.port = 443 if use_ssl else 80
        else:
            self.port = port

        self.connected = False
        # Session is useful because it does not add "Connection: close header"
        # Since we will be making same call to the same URL, making a session
        # REF: https://requests.readthedocs.io/en/latest/user/advanced/
        self.http_session = requests.Session()

        self.last_response   = ''
        self.last_query_time = 0

        self.url = ('https' if self.use_ssl else 'http') + \
            '://' + host + ':' + str(port) + '/api/'

        self.token = token

    def __del__(self):
        logger.info("Done with connector REST.")
        self.http_session.close()

    def _query(self, query, blob_array = [], try_resume=True):
        response_blob_array = []
        # Check the query type
        if not isinstance(query, str):  # assumes json
            query_str = json.dumps(query)
        else:
            query_str = query

        files = [
            ('query', (None, query_str)),
        ]

        for blob in blob_array:
            files.append(('blobs', blob))

        # Set Auth token, only when not authenticated before
        if self.shared_data.session and self.shared_data.session.valid():
            headers = {'Authorization': "Bearer " +
                       self.shared_data.session.session_token}
        else:
            headers = None
        tries = 0
        response = SimpleNamespace()
        response.status_code = 0
        while tries < self.config.retry_max_attempts:
            tries += 1
            try:
                response = self.http_session.post(self.url,
                                                  headers = headers,
                                                  files   = files,
                                                  verify  = self.use_ssl)
                if response.status_code == 200:
                    # Parse response:
                    json_response       = json.loads(response.text)
                    import base64
                    response_blob_array = [base64.b64decode(
                        b) for b in json_response['blobs']]
                    self.last_response  = json_response["json"]
                    break
                logger.error(
                    f"Response not OK = {response.status_code} {response.text[:1000]}\n\
                        attempt [{tries}/3] .. PID = {os.getpid()}")
            except Exception as e:
                logger.error(f"Exception during http.post = {e}\n\
                        attempt [{tries}/3] .. PID = {os.getpid()}")

            time.sleep(self.config.retry_interval_seconds)

            self.connect()
            if try_resume:
                self._renew_session()

        if tries == self.config.retry_max_attempts:
            raise Exception(
                f"Could not query ApertureDB {self.config} using REST.")
        return (self.last_response, response_blob_array)

    def _connect(self):
        logger.info("Connecting to ApertureDB using REST")
        self.connected = True



#!/usr/bin/env python3

"""
This file contains some utilities to help with converting existing CSV files
to the format required by ApertureDB.  The main functions are for writing
CSV files for creating entities, images, and connections.  The functions
take in a pandas DataFrame and write it to a CSV file with the appropriate
headers required by the various *CSV loaders in ApertureDB.  This makes it
easier to generate CSV files without having to remember the exact column
names required by the CSV loaders.
"""

from typing import Optional

import pandas as pd


def convert_entity_data(input, entity_class: str, unique_key: Optional[str] = None):
    """
    Convert data to the format required for creating entities

    Arguments:
        input: Anything that can be used as input to a pandas DataFrame, including a pandas DataFrame
        entity_class: The entity class to write to the CSV file as the first column
        unique_key: (optional) An existing key to call out as a constraint

    Returns:
        A pandas DataFrame with the entity class as the first column
    """
    df = pd.DataFrame(input)
    df.insert(0, 'EntityClass', entity_class)
    if unique_key:
        assert unique_key in df.columns, f"unique_key {unique_key} not found in the input data"
        df[f"constraint_{unique_key}"] = df[unique_key]
    return df


def write_entity_csv(filename: str, input, **kwargs):
    """
    Write data to a CSV file for creating entities

    Arguments:
        filename: The name of the file to write to.
            Recommended to end in ".entity.csv".
        input: Anything that can be used as input to a pandas DataFrame, including a pandas DataFrame
        entity_class: The entity class to write to the CSV file as the first column
        unique_key: (optional) An existing key to call out as a constraint
    """
    df = convert_entity_data(input, **kwargs)
    df.to_csv(filename, index=False)


def convert_image_data(input, source_column: str, source_type: Optional[str] = None, format: Optional[str] = None, unique_key: Optional[str] = None):
    """
    Convert data to the format required for creating images

    Arguments:
        input: Anything that can be used as input to a pandas DataFrame, including a pandas DataFrame
        source_column: The name of the column that contains the image data
        source_type: (optional) The type the source column.  If not specified, the source column will be used.  Should be one of "filename", "url", "gsurl", or "s3url".
        format: (optional) The format of the image data.  If not provided, there should be a column called "format" in the input data.
        unique_key: (optional) An existing key to call out as a constraint

    Returns:
        A pandas DataFrame with the source column as the first column
    """
    df = pd.DataFrame(input)

    assert source_column in df.columns, f"source_column {source_column} not found in the input data"

    if source_type is None:
        source_type = source_column

    assert source_type in ["filename", "url", "gsurl",
                           "s3url"], f"source_type must be one of 'filename', 'url', 'gsurl', or 's3url', found: {source_type}"

    if source_column == source_type:
        # reordering the columns to make the source column the first column
        df = df[[source_column] +
                [col for col in df.columns if col != source_column]]
    else:
        df.insert(0, source_type, df[source_column])

    if unique_key is not None:
        assert unique_key in df.columns, f"unique_key {unique_key} not found in the input data"
        df[f"constraint_{unique_key}"] = df[unique_key]

    if format is not None:
        assert 'format' not in df.columns, "format column already exists in the input data"
        df['format'] = format
    else:
        assert 'format' in df.columns, "format column not found in the input data"
        # Reorder the columns to make the format column the last column
        df = df[[col for col in df.columns if col != 'format'] + ['format']]

    return df


def write_image_csv(filename: str, input, **kwargs):
    """
    Write data to a CSV file for creating images

    Arguments:
        filename: The name of the file to write to.
            Recommended to end in ".image.csv".
        input: Anything that can be used as input to a pandas DataFrame, including a pandas DataFrame
        source_column: The name of the column that contains the image data
        source_type: (optional) The type the source column.  If not specified, the source column will be used.  Should be one of "filename", "url", "gsurl", or "s3url".
        format: (optional) The format of the image data.  If not provided, there should be a column called "format" in the input data.
        unique_key: (optional) An existing key to call out as a constraint
    """
    df = convert_image_data(input, **kwargs)
    df.to_csv(filename, index=False)


def convert_connection_data(input,
                            connection_class: str,
                            source_class: str, source_property: str,
                            destination_class: str, destination_property: str,
                            source_column: Optional[str] = None, destination_column: Optional[str] = None,
                            unique_key: Optional[str] = None):
    """
    Convert data to the format required for creating connections

    Arguments:
        input: Anything that can be used as input to a pandas DataFrame, including a pandas DataFrame
        connection_class: The connection class to write to the CSV file as the first column
        source_class: The source entity class
        source_property: The property containing the source key in ApertureDB
        destination_class: The destination entity class
        destination_property: The property of the destination entity in ApertureDB
        source_column: (optional) The column containing the source keys in the input data.  Defaults to the source_property.
        destination_column: (optional) The column containing the destination keys in the input data.  Defaults to the destination_property.
        unique_key: (optional) An existing key to call out as a constraint
        """
    assert "@" not in source_class, "source_class should not contain '@'"
    assert "@" not in destination_class, "destination_class should not contain '@'"

    df = pd.DataFrame(input)

    if source_column is None:
        source_column = source_property
    assert source_column in df.columns, f"source_column {source_column} not found in the input data"

    if destination_column is None:
        destination_column = destination_property
    assert destination_column in df.columns, f"destination_column {destination_column} not found in the input data"

    df.insert(0, 'ConnectionClass', connection_class)
    df.insert(1, f"{source_class}@{source_property}", df[source_column])
    df.insert(2, f"{destination_class}@{destination_property}",
              df[destination_column])

    if unique_key:
        assert unique_key in df.columns, f"unique_key {unique_key} not found in the input data"
        df[f"constraint_{unique_key}"] = df[unique_key]

    return df


def write_connection_csv(filename: str, input, **kwargs):
    """
    Write data to a CSV file for creating connections

    Arguments:
        filename: The name of the file to write to.
            Recommended to end in ".connection.csv".
        input: Anything that can be used as input to a pandas DataFrame, including a pandas DataFrame
        connection_class: The connection class to write to the CSV file as the first column
        source_class: The source entity class
        source_property: The property containing the source key in ApertureDB
        source_column: (optional) The column containing the source keys in the input data.  Defaults to the source_property.
        destination_class: The destination entity class
        destination_property: The property of the destination entity in ApertureDB
        destination_column: (optional) The column containing the destination keys in the input data.  Defaults to the destination_property.
        unique_key: (optional) An existing key to call out as a constraint
    """
    df = convert_connection_data(input, **kwargs)
    df.to_csv(filename, index=False)



from aperturedb.Entities import Entities


class Clips(Entities):
    """
    **The object mapper representation of Video Clips in ApertureDB.**

    This class is a layer on top of the native query.
    It facilitates interactions with Video clips in the database in the pythonic way.
    """
    db_object = "_Clip"



# queryMessage.py - wraps protobuf versions
import google.protobuf

if google.protobuf.__version__.split(".")[0] == "3":
    from . import queryMessage3_pb2

    def queryMessage():
        return queryMessage3_pb2.queryMessage()

    def ParseFromString(msg, data):
        return msg.ParseFromString(data)
elif google.protobuf.__version__.split(".")[0] == "4":
    from . import queryMessage4_pb2

    def queryMessage():
        return queryMessage4_pb2.queryMessage()

    def ParseFromString(msg, data):
        # because of https://github.com/protocolbuffers/protobuf/issues/10774
        return msg.ParseFromString(memoryview(data).tobytes())
else:
    raise Exception(
        f"aperturedb not compatible with {google.protobuf.__version__}")



from __future__ import annotations

from enum import Enum
from typing import List, Tuple, Union

from aperturedb.Constraints import Constraints
from aperturedb.Operations import Operations
from aperturedb.Sort import Sort
from pydantic import BaseModel
from aperturedb.Sources import Sources
import struct


class ObjectType(Enum):
    """
    Based on : [API description](/query_language/Overview/API%20Description)

    """
    BLOB = "_Blob"
    BOUNDING_BOX = "_BoundingBox"
    CONNECTION = "_Connection"
    CLIP = "_Clip"
    DESCRIPTOR = "_Descriptor"
    DESCRIPTORSET = "_DescriptorSet"
    ENTITY = "_Entity"
    FRAME = "_Frame"
    IMAGE = "_Image"
    POLYGON = "_Polygon"
    VIDEO = "_Video"


class Config(Enum):
    SAVE_NAME = 1
    SAVE_VALUE = 2


class PropertyType(str, Enum):
    SYSTEM = "system"
    USER = "user"


class RangeType(str, Enum):
    TIME = "time"
    FRAME = "frame"
    FRACTION = "fraction"


config = Config.SAVE_NAME


def class_entity(db_class: Union[str, ObjectType]):
    members = [m.value for m in ObjectType]
    db_class = db_class if not isinstance(
        db_class, ObjectType) else db_class.value
    if db_class.startswith("_"):
        if db_class in members:
            return f"{db_class[1:]}"
        else:
            raise Exception(
                f"Invalid Object type. Should not begin with _, except for {members}")
    else:
        return "Entity"


def get_handlers():
    sources = Sources(n_download_retries=3)
    source_url_handlers = {
        "http": sources.load_from_http_url,
        "https": sources.load_from_http_url,
        "": sources.load_from_file,
        "s3": sources.load_from_s3_url,
        "gs": sources.load_from_gs_url
    }
    return source_url_handlers


def get_specific(obj: BaseModel) -> dict:
    """This function is used to get the specific parameters for the object.
    Example: For a Descriptor object, the vector is packed into a blob.
    """
    if obj.type == ObjectType.CLIP:
        range_types = {
            RangeType.TIME: "time_offset_range",
            RangeType.FRAME: "frame_number_range",
            RangeType.FRACTION: "time_fraction_range"
        }
        start, stop  = obj.start, obj.stop
        if obj.range_type == RangeType.TIME:
            start, stop = int(start), int(stop)
            start = f"{start//60}:{start%60}"
            stop = f"{stop//60}:{stop%60}"
        elif obj.range_type == RangeType.FRAME:
            start = int(obj.start)
            stop = int(obj.stop)
        return{
            range_types[obj.range_type]: {
                "start": start,
                "stop": stop,
            }
        }, []
    elif obj.type == ObjectType.DESCRIPTOR:
        blob = struct.pack("%sf" % len(obj.vector), *obj.vector)
        return {
            "set": obj.set.name
        }, [blob]
    elif obj.type == ObjectType.DESCRIPTORSET:
        return {
            "name": obj.name,
            "dimensions": obj.dimensions
        }, []
    return {}, []


def generate_add_query(
        obj: BaseModel,
        cached: List[str] = None,
        source_field: str = None,
        index: int = 1,
        parent: int = 0) -> Tuple[List[object], List[bytearray], int]:
    """
    Takes the user model, and builds out a sequence of commands that creates
    a similar structure on apertureDB's graph. For example, given the following

    ```mermaid
    erDiagram
          VIDEO {
            CLIP range
          }
          CLIP{
            DESCRIPTOR vector
          }
          VIDEO ||--o{ CLIP : clips
          CLIP ||--|| DESCRIPTOR : descriptor
    ```

    The function will generate the following commands:

    *AddVideo 1*
        *AddClip 2*
        *AddDescriptor 3*
        *AddConnection 3-2*

    .
    .
    .

    Args:
        obj (BaseModel): The object from the user domain.
        cached (List[str], optional): helps to optimize sending one blob per node. Defaults to None.
        source_field (str, optional): Preserves the relevant connection information from user objects. Defaults to None.
        index (int, optional): The index to start creating references from. Defaults to 1.
        parent (int, optional): The parent of the current node. Defaults to None.

    Returns:
        tuple : Contains 3 items:
            * List of commands with current node and its children
            * List of blobs with current node and its children
            * index, adjusting the count of children + self for the number of commands
              in current query
    """
    source_url_handlers = get_handlers()
    if cached == None:
        cached = []
    query = []
    dependents = []
    dependents_blobs = []
    props = {}
    blobs = []
    cindex = index
    specific_params = {}
    specific_blobs = []
    if obj.id not in cached:
        for p in obj.__dict__.keys():
            if "_" == p[0]:
                continue
            subobj = getattr(obj, p)
            meta = obj.model_fields[p].metadata
            if len(meta) > 0 and meta[0] == PropertyType.SYSTEM:
                specific_params, specific_blobs = get_specific(obj)
            elif isinstance(subobj, int) or isinstance(subobj, str) or isinstance(subobj, str) or isinstance(subobj, float):
                props[p] = subobj
            elif isinstance(subobj, Enum):
                props[p] = subobj.name if config == Config.SAVE_NAME else subobj.value
            elif "vector" == p:
                pass
            elif isinstance(subobj, list):
                for i, si in enumerate(subobj):
                    q, b, index = generate_add_query(
                        si, cached=cached, source_field=f"{p}[{i}]", index=index + 1, parent=cindex)
                    dependents_blobs.extend(b)
                    dependents.extend(q)
            else:
                q, b, index = generate_add_query(
                    subobj, cached=cached, source_field=f"{p}", index=index + 1, parent=cindex)
                dependents.extend(q)
                dependents_blobs.extend(b)

    # We still want to create dummy Add* commands so that references
    # for creating already created instances are mantained.
    params = {
        "_ref": cindex,
        "properties": props if "id" in props else {},
    }
    if obj.type != ObjectType.DESCRIPTORSET:
        params["if_not_found"] = {
            "id": ["==", props["id"] if "id" in props else obj.id]
        }
    for k, v in specific_params.items():
        params[k] = v
    blobs.extend(specific_blobs)
    if hasattr(obj, "type"):
        props.pop('type', None)
        if obj.type != ObjectType.ENTITY:
            props["user_type"] = type(obj).__name__
            # Generates a Add<ADB object> command
            if obj.id not in cached:
                query.append(
                    QueryBuilder.add_command(obj.type.value, params=params))
            else:
                if "if_not_found" in params:
                    params["constraints"] = params["if_not_found"]
                    params.pop("if_not_found", None)
                params.pop("properties", None)
                query.append(
                    QueryBuilder.find_command(obj.type.value, params=params))
        if obj.type in [ObjectType.IMAGE, ObjectType.VIDEO, ObjectType.BLOB]:
            # Do not send blob, if Node has been added to set of commands.
            if obj.id not in cached:
                if obj.url:
                    parts = obj.url.split(":")
                    if len(parts) > 1:
                        _, blob = source_url_handlers[parts[0]](
                            obj.url, lambda x: True)
                        blobs.append(blob)
                    else:
                        _, blob = source_url_handlers[""](obj.url)
                        blobs.append(blob)
                props.pop('url', None)
        else:
            # Generates an AddEntity
            if obj.type == ObjectType.ENTITY:
                query.append(
                    QueryBuilder.add_command(type(obj).__name__, params=params))

        cached.append(obj.id)

    if parent > 0:
        if obj.type == ObjectType.CLIP:
            query[0]["AddClip"]["video_ref"] = parent

        else:
            params = {
                "src": parent,
                "dst": cindex,
                "class": type(obj).__name__,
                "properties": {
                    "source_field": source_field
                }
            }
            query.append(
                QueryBuilder.add_command(
                    ObjectType.CONNECTION.value, params=params)
            )
    query.extend(dependents)
    blobs.extend(dependents_blobs)
    return query, blobs, index


class QueryBuilder():
    @classmethod
    def find_command(self, oclass: Union[str, ObjectType], params: dict) -> dict:
        return self.build_command(oclass, params, "Find")

    @classmethod
    def add_command(self, oclass: Union[str, ObjectType], params: dict) -> dict:
        return self.build_command(oclass, params, "Add")

    @classmethod
    def build_command(self, oclass: Union[str, ObjectType], params: dict, operation: str) -> dict:
        oclass = oclass if not isinstance(oclass, ObjectType) else oclass.value
        command = {
            f"{operation}Entity": params
        }
        members = [m.value for m in ObjectType]
        if oclass.startswith("_"):
            if oclass in members:
                command = {
                    f"{operation}{oclass[1:]}": params
                }
            else:
                raise Exception(
                    f"Invalid Object type. Should not begin with _, except for {members}")
        else:
            if operation == "Find":
                params["with_class"] = oclass
            else:
                params["class"] = oclass
        return command


class Query():
    """
    This is the underlying class to generate a query using python code.
    """
    db_object = "Entity"
    next = None

    def connected_to(self,
                     spec: Query,
                     adj_to: int = 0) -> Query:
        """
        Can be used to connect two commands together, for non trivial queries.
        This is the lower level function to create a query.

        Args:
            spec (Query): One command of the query.
            adj_to (int, optional): The command to refer to. Defaults to 0.

        Returns:
            Query: The completed, connected query.
        """
        spec.adj_to = self.adj_to + 1 if adj_to == 0 else adj_to
        self.next = spec
        return self

    def command_properties(self, prop: str = "") -> List[str]:
        """
        Helper function to get the properties of all commands the query.

        Args:
            v (str, optional): _description_. Defaults to "".

        Returns:
            List[str]: Properties from commands in the order they should be executed.
        """
        chain = []
        p = self
        while p is not None:
            chain.append(getattr(p, prop))
            p = p.next
        return chain

    @classmethod
    @classmethod
    def spec(cls,
             constraints: Constraints = None,
             operations: Operations = None,
             with_class: str = "",
             limit: int = -1,
             sort: Sort = None,
             list: List[str] = None,
             group_by_src: bool = False,
             blobs: bool = False,
             set: str = None,
             vector: List[float] = None,
             k_neighbors: int = 0
             ) -> Query:
        """
        The [specification](/query_language/Overview/API%20Description) for a command to be used in a query.

        Args:
            constraints (Constraints, optional): [Constraints](/query_language/Reference/shared_command_parameters/constraints) . Defaults to None.
            with_class (ObjectType, optional): _description_. Defaults to ObjectType.CUSTOM.
            limit (int, optional): _description_. Defaults to -1.
            sort (Sort, optional): _description_. Defaults to None.
            list (List[str], optional): _description_. Defaults to None.

        Returns:
            Query: The query object.
        """
        return Query(
            constraints=constraints,
            operations=operations,
            with_class=with_class,
            limit=limit,
            sort = sort,
            list = list,
            blobs=blobs,
            group_by_src = group_by_src,
            set=set,
            vector=vector,
            k_neighbors=k_neighbors
        )

    def __init__(self,
                 constraints: Constraints = None,
                 operations: Operations = None,
                 with_class: str = "",
                 limit: int = -1,
                 sort: Sort = None,
                 list: List[str] = None,
                 group_by_src: bool = False,
                 blobs: bool = False,
                 adj_to: int = 0,
                 set: str = None,
                 vector: List[float] = None,
                 k_neighbors: int = 0
                 ):
        self.constraints = constraints
        self.operations = operations
        self.with_class = with_class
        self.limit = limit
        self.sort = sort
        self.list = list
        self.group_by_src = group_by_src
        self.blobs = blobs
        self.adj_to = adj_to + 1
        self.set = set
        self.vector = vector
        self.k_neighbors = k_neighbors

    def query(self) -> List[dict]:
        results_section = "results"
        cmd_params = {results_section: {}, "_ref": self.adj_to}
        if self.limit != -1:
            cmd_params[results_section]["limit"] = self.limit
        if self.sort:
            cmd_params[results_section]["sort"] = self.sort._sort
        if self.list is not None and len(self.list) > 0:
            cmd_params[results_section]["list"] = self.list
        else:
            cmd_params[results_section]["all_properties"] = True
        cmd_params[results_section]["group_by_source"] = self.group_by_src

        if self.constraints:
            cmd_params["constraints"] = self.constraints.constraints
        if self.operations:
            cmd_params["operations"] = self.operations.operations_arr
        if self.set:
            cmd_params["set"] = self.set
        if self.k_neighbors > 0:
            cmd_params["k_neighbors"] = self.k_neighbors

        self.blob = None
        if self.vector:
            self.blob = struct.pack("%sf" % len(self.vector), *self.vector)

        self.with_class = self.with_class if self.db_object == "Entity" else \
            self.db_object if not isinstance(
                self.db_object, ObjectType) else self.db_object.value
        cmd = QueryBuilder.find_command(
            oclass=self.with_class, params=cmd_params)
        self.find_command = list(cmd.keys())[0]
        query = [cmd]
        blobs = []
        if self.blob:
            blobs.append(self.blob)

        if self.next:
            next_commands, next_blobs = self.next.query()
            list(next_commands[0].values())[0]["is_connected_to"] = {
                "ref": self.adj_to
            }
            # list(next_commands[0].values())[0]["_ref"] = self.adj_to
            query.extend(next_commands)
            blobs.extend(next_blobs)

        return query, blobs



import logging
from aperturedb.Blobs import Blobs
from aperturedb.BoundingBoxes import BoundingBoxes
from aperturedb.Constraints import Constraints
from aperturedb.Entities import Entities
from aperturedb.Images import Images
from aperturedb.Query import Query
from aperturedb.Query import generate_add_query
from aperturedb.DataModels import ImageDataModel, IdentityDataModel
import random
from typing import List
from enum import Enum

logger = logging.getLogger(__file__)


class TestDataWizardEntities():
    def test_get_persons(self, insert_data_from_csv, db):
        loaded, _ = insert_data_from_csv(in_csv_file="./input/persons.adb.csv")
        all_persons = Entities.retrieve(db,
                                        spec=Query.spec(with_class="Person"))
        assert len(all_persons) == len(loaded)
        too_young = all_persons.filter(lambda p: p["age"] < 18)
        too_old = all_persons.filter(lambda p: p["age"] > 60)
        working_people = all_persons - too_old - too_young

        assert all([p["age"] >= 18 and p["age"] <= 60 for p in working_people])
        ordered = working_people.sort(key=lambda p: p["age"])
        youngest_working_five = ordered[:5]
        eldest_working_five = ordered[-5:]

        assert len(youngest_working_five) == 5
        assert len(eldest_working_five) == 5
        assert youngest_working_five[0]["age"] < eldest_working_five[0]["age"]

    def test_get_persons_constraints(self, retired_persons):
        logger.info(f"Retired persons count = {len(retired_persons)}")
        assert all(p["age"] >= 60 for p in retired_persons)

    def test_get_persons_display(self, retired_persons):
        logger.info(f"Retired persons count = {len(retired_persons)}")
        df = retired_persons.inspect()
        logger.info(f"\n{df}")
        ages = df['age']
        assert ages[ages >= 60].count() == len(df)

    def test_update_properties(self, db, retired_persons):
        risk_factors = [{
            "risk_factor": (p["age"] - 60) // 10
        } for p in retired_persons]
        retired_persons.update_properties(risk_factors)
        df = retired_persons.inspect()
        logger.info(f"\n{df}")
        persons = Entities.retrieve(db,
                                    spec=Query.spec(with_class="Person",
                                                    constraints=Constraints().greaterequal("age", 60)))

        assert len(persons) == len(retired_persons)
        assert all(p["risk_factor"] == (p["age"] - 60) // 10 for p in persons)


class TestDataWizardImages():
    def test_get_images(self, insert_data_from_csv, db):
        loaded, _ = insert_data_from_csv(
            "./input/images.adb.csv", rec_count=10)
        images = Images.retrieve(db, Query.spec(limit=10))
        assert isinstance(
            images, Images), f"{type(images)} is not an instance of Images"
        logger.info(f"\n{images.inspect()}")
        assert images != None
        assert len(images) <= 10


class TestDataWizardBlobs():
    def test_get_blobs(self, insert_data_from_csv, db):
        loaded, _ = insert_data_from_csv("./input/blobs.adb.csv", rec_count=10)
        blobs = Blobs.retrieve(db, Query.spec(limit=10))
        assert isinstance(
            blobs, Blobs), f"blobs is not an instance of Blobs = {blobs}"
        logger.info(f"\n{blobs.inspect()}")
        assert blobs != None
        assert len(blobs) <= 10


class TestDataWizardBoundingBoxes():
    def test_get_bounding_boxes(self, insert_data_from_csv, db):
        loaded, _ = insert_data_from_csv(
            "./input/images.adb.csv", rec_count=10)
        loaded, _ = insert_data_from_csv(
            "./input/bboxes.adb.csv", rec_count=10)
        bboxes = BoundingBoxes.retrieve(db, Query.spec(limit=10))
        assert isinstance(
            bboxes, BoundingBoxes), f"bboxes is not an instance of BoundingBoxes = {bboxes}"
        logger.info(f"\n{bboxes.inspect()}")
        assert bboxes != None
        assert len(bboxes) <= 10


def make_people(count: int = 1) -> List[object]:
    class Side(Enum):
        RIGHT = 1
        LEFT = 2

    class Finger(IdentityDataModel):
        nail_clean: bool = False

    class Hand(ImageDataModel):
        side: Side = None
        thumb: Finger = None
        fingers: List[Finger] = []

    class Person(IdentityDataModel):
        name: str = ""
        hands: List[Hand] = []
        dominant_hand: Hand = None

    def make_hand(side: Side) -> Hand:
        hand = Hand(side = side, url= "input/images/0079.jpg")
        hand.fingers = [Finger(nail_clean=True) if random.randint(
            0, 1) == 1 else Finger(nail_clean=False) for i in range(5)]
        hand.thumb = hand.fingers[0]
        return hand

    people = []
    for i in range(10):
        person = Person(name=f"adam{i+1}")
        left_hand = make_hand(Side.LEFT)
        right_hand = make_hand(Side.RIGHT)
        person.hands.extend([left_hand, right_hand])
        person.dominant_hand = person.hands[0]
        people.append(person)
    return people


class TestQueryBuilder():
    def test_all_info_preserved(self):
        # this will create 10 people entities.
        # Each person will have 2 hands (+1 connection via dominant hand)
        # Each hand will have 10 fingers (+2 connections via thumb)
        # Everything will be interconnected.
        # Nodes per person = 1 + 3 + 12 = 16 (The commands will be generated with if not found)
        # Connections per person = 3 for hand 12 for fingers
        # Commands = 16 + 15 = 31
        people = make_people(10)
        total_commands = []
        total_blobs = []
        for person in people:
            q, b, current_ref = generate_add_query(person)
            total_commands += q
            total_blobs += b
            # 16 entities have been inserted and referenced
            assert current_ref == 16

        assert len(total_commands) == 310
        assert len(total_blobs) == 20
        assert len(
            list(filter(lambda cmd: "AddConnection" in cmd, total_commands))) == 150
        assert len(
            list(filter(lambda cmd: "AddEntity" in cmd, total_commands))) == 130
        assert len(
            list(filter(lambda cmd: "AddImage" in cmd, total_commands))) == 20
        assert len(
            list(filter(lambda cmd: "FindImage" in cmd, total_commands))) == 10



# SPARQL wrapper for ApertureDB
#
# This script allows ApertureDB to be queried using SPARQL queries.
# It is based on the rdflib library.
# Currently, it supports a subset of BGP queries, but it can be extended to support more complex queries.

from typing import Dict, Generator, List, Tuple, Union, Optional
from urllib.parse import quote, unquote
import json
import math
import logging

from aperturedb.CommonLibrary import create_connector, execute_query
from aperturedb.Utils import Utils


class SPARQL:
    def __init__(self, client=None, debug=False, log_level=None):
        """SPARQL compatability layer for ApertureDB

        Args:
            client: ApertureDB client. If not suppliued, then `create_connector()` is used to create a new client.
            debug: bool. If True, then certain intermediate results are stored in the object.
            log_level: int. The logging level to use for the ApertureDB client.
        """
        self.debug = debug
        self.logger = logging.getLogger(__name__)
        if log_level is not None:
            self.logger.setLevel(log_level)

        try:
            import rdflib
        except ImportError:
            raise ImportError(
                "The rdflib library is required to use the SPARQL interface. "
                "Please install it using 'pip install rdflib'."
            )

        if client is None:
            client = create_connector()
        self._client = client
        self._utils = Utils(client)
        namespace = rdflib.Namespace("http://aperturedb.io/")
        self.namespaces = {
            "t": namespace["type/"],
            "c": namespace["connection/"],
            "p": namespace["property/"],
            "o": namespace["object/"],
            "knn": namespace["knn/"],
        }
        self._load_schema()
        # TODO: Only one instance of SPARQL can be used at a time
        import rdflib.plugins.sparql
        rdflib.plugins.sparql.CUSTOM_EVALS["aperturedb"] = self.eval
        self.graph = rdflib.Graph()
        for k, v in self.namespaces.items():
            self.graph.bind(k, v)
        self.knn_properties = set(self.namespaces["knn"] + p for p in [
                                  "similarTo", "set", "vector", "k_neighbors", "knn_first", "distance", "engine", "metric"])

        self.logger.info(
            f"SPARQL loaded {len(self.connections)} connections and {len(self.properties)} properties")

    def __del__(self):
        from rdflib.plugins.sparql import CUSTOM_EVALS
        if CUSTOM_EVALS.get("aperturedb") == self.eval:
            del CUSTOM_EVALS["aperturedb"]

    def _make_uri(self, prefix, suffix):
        return self.namespaces[prefix] + quote(suffix)

    def _parse_uri(self, uri: "URIRef"):
        from rdflib.term import URIRef
        assert isinstance(uri, URIRef)
        for k, v in self.namespaces.items():
            if uri.startswith(v):
                return k, unquote(uri[len(v):])
        return None, None

    def _parse_uri_with_ns(self, ns: str, uri: "URIRef"):
        from rdflib.term import URIRef
        assert isinstance(uri, URIRef), uri
        assert uri.startswith(self.namespaces[ns]), (ns, uri)
        return unquote(uri[len(self.namespaces[ns]):])

    def _format_node(self, node):
        from rdflib.term import URIRef
        return self.graph.qname(node) if isinstance(node, URIRef) else node.toPython()

    def _format_triple(self, triple):
        return " ".join([self._format_node(node) for node in triple])

    def _format_triples(self, triples):
        return " .\n".join([self._format_triple(triple) for triple in triples])

    def _load_schema(self):
        self.schema = self._utils.get_schema()
        self.connections = {}
        self.connections = {}
        if self.schema is None:
            self.logger.warning("No schema found in ApertureDB")
            return
        if "connections" in self.schema and self.schema["connections"] is not None and "classes" in self.schema["connections"]:
            for c, d in self.schema["connections"]["classes"].items():
                uri = self._make_uri("c", c)
                if uri not in self.connections:
                    self.connections[uri] = (set(), set())
                self.connections[uri][0].add(d["src"])
                self.connections[uri][1].add(d["dst"])
        if not self.connections:
            self.logger.warning("No connections found in schema")

        self.properties = {}
        if "entities" in self.schema and self.schema["entities"] is not None and "classes" in self.schema["entities"]:
            for e, d in self.schema["entities"]["classes"].items():
                for p in d["properties"]:
                    uri = self._make_uri("p", p)
                    if uri not in self.properties:
                        self.properties[uri] = set()
                    self.properties[uri].add(e)
                self.namespaces[f"{e}"] = self._make_uri("o", e + "/")
        if not self.properties:
            self.logger.warning("No properties found in schema")

    def eval(self, ctx: "QueryContext", part: "CompValue"
             ) -> Generator["FrozenBindings", None, None]:
        """
        Evaluate a SPARQL query part

        This method is registered as a custom eval function for rdflib.
        It is given the first opportunity to evaluate a query part.

        Arguments:
            ctx: QueryContext. The query context, supplying existing bindings
            part: CompValue. The query part to evaluate

        Yields:
            FrozenBindings. The results of the query

        Raises:
            NotImplementedError: If the query type is not supported.
                This is the case for all query types except BGP.
        """
        if part.name == 'BGP':
            return self.evalBGP(ctx, part.triples)

        raise NotImplementedError(f"Unsupported query type: {part.name}")

    def evalBGP(self, ctx: "QueryContext",
                triples: List[Tuple["Identifier", "Identifier", "Identifier"]],
                ) -> Generator["FrozenBindings", None, None]:
        """
        Evaluates a Basic Graph Pattern (BGP) query

        Arguments:
            ctx: QueryContext. The query context, supplying existing bindings
            triples: List[Tuple[Identifier, Identifier, Identifier]]. The triples to evaluate

        Yields:
            FrozenBindings. Binding sets that are the results of the query
        """
        def add_find(v, t):
            """Create new Find* command for variable v with type t"""
            from rdflib.term import Variable
            command_name = "FindEntity" if t[0] != "_" else "Find" + t[1:]
            body = {}
            if t[0] != "_":
                body["with_class"] = t
            body["_ref"] = len(query) + 1
            body["uniqueids"] = True
            commands[v] = body
            command = {command_name: body}
            query.append(command)
            if isinstance(v, Variable):
                # print("Adding uniqueid binding for variable", v)
                bindings.append((v, body, "_uniqueid"))
            object_prefixes.append(self._make_uri("o", t + "/"))

        def add_constraint(s, p, o):
            """Add property constraint to existing command"""
            command = commands[s]
            ns, prop = self._parse_uri(p)
            assert ns in ["p", "knn"], (ns, p)
            if "constraints" not in command:
                command["constraints"] = {}
            if prop not in command["constraints"]:
                command["constraints"][prop] = []
            command["constraints"][prop].extend(["==", o.toPython()])

        def add_binding(s, p, o):
            """Associates a return binding variable with a specific property in the ApertureDB response"""
            command = commands[s]
            if "results" not in command:
                command["results"] = {}
            if "list" not in command["results"]:
                command["results"]["list"] = []
            prop = self._parse_uri_with_ns("p", p)
            command["results"]["list"].append(prop)
            bindings.append((o, command, prop))

        def add_is_connected_to(s, p, o):
            """Adds connection constraint between two commands"""
            s_command = commands[s]
            o_command = commands[o]
            if s_command["_ref"] < o_command["_ref"]:
                command = o_command
                ict = dict(
                    ref=s_command["_ref"],
                    direction="out",
                )
                connection_bindings.append((o_command, p, s_command, False))
            else:
                command = s_command
                ict = dict(
                    ref=o_command["_ref"],
                    direction="in",
                )
                connection_bindings.append((s_command, p, o_command, True))

            if p is not None:
                connection_type = self._parse_uri_with_ns("c", p)
                ict["connection_class"] = connection_type
            else:
                del ict["direction"]  # Any connection, either in or out

            if "is_connected_to" not in command:
                command["is_connected_to"] = ict
                command["group_by_source"] = True
            elif list(command["is_connected_to"].keys())[0] == "all":
                list(command["is_connected_to"].values())[0].append(ict)
            else:
                command["is_connected_to"] = {
                    "all": [command["is_connected_to"], ict]}

        def process_bindings(output, ctx: "QueryContext" = ctx, i=0, ids=set()):
            """Processes AperatureDB response and yields bindings

            Operates recursively, processing each command in the query to generate bindings
            for each variable, multiplying the number of bindings at each step.
            Uses `solutions` to avoid yielding duplicate solutions.
            """
            from rdflib.term import Variable, BNode, Literal, URIRef
            if i == len(output):
                solution = ctx.solution()
                # print("Yielding solution", solution)
                if solution in solutions:
                    return
                solutions.add(solution)
                yield solution
                return

            command_body = list(query[i].values())[0]
            result_body = list(output[i].values())[0]
            entities = result_body.get("entities", [])

            bindings2 = [(v, prop) for v, cb, prop in bindings
                         if cb == command_body and not isinstance(v, BNode)]
            assert all(isinstance(v, Variable)
                       for v, _ in bindings2), bindings2
            assert all(ctx[v] is None for v, _ in bindings2), bindings2

            if isinstance(entities, list):
                entities = {None: entities}

            for id_, ee in entities.items():
                if id_ is not None and id_ not in ids:
                    continue
                for e in ee:
                    assert "_uniqueid" in e, e
                    id2 = e["_uniqueid"]
                    ids2 = ids.union({id2})
                    ctx2 = ctx.push()  # not always necessary
                    if bindings2:
                        for v, prop in bindings2:
                            assert prop in e, (prop, e)
                            assert ctx2[v] is None, (v, ctx2[v])
                            if prop == "_uniqueid":
                                uri = URIRef(
                                    object_prefixes[i] + e["_uniqueid"])
                                ctx2[v] = uri
                            else:
                                literal = Literal(e[prop])
                                ctx2[v] = literal
                        yield from process_bindings(output, ctx2, i + 1, ids2)
                    else:
                        yield from process_bindings(output, ctx, i + 1, ids2)

        from rdflib import RDF, Literal
        from rdflib.plugins.sparql.sparql import SPARQLError

        triples = self._apply_context_to_triples(ctx, triples)
        types = self._deduce_types(ctx, triples)
        query = []  # ApertureDB query
        commands = {}  # Mapping from variable to command body
        bindings = []  # List of (variable, command body, property) tuples
        object_prefixes = []  # List of object prefixes
        # List of (source command, connection, destination command, is_forwards) tuples
        connection_bindings = []
        blobs = []  # List of blobs

        # Add Find commands for each variable
        for v, t in types.items():
            add_find(v, t)
        query, object_prefixes = self._optimize_query(query, object_prefixes)

        # Find knn nodes
        knn_nodes = {
            o: s for s, p, o in triples
            if p == self.namespaces["knn"] + "similarTo"}

        # Process triples to generate querty
        # Add constraints, bindings, connections
        for s, p, o in triples:
            if p in self.properties:
                if isinstance(o, Literal):
                    add_constraint(s, p, o)
                else:
                    add_binding(s, p, o)
            elif p in self.connections:
                add_is_connected_to(s, p, o)
            elif p == RDF.type:
                pass
            elif p in self.knn_properties:
                if p == self.namespaces["knn"] + "similarTo":
                    pass
                elif p == self.namespaces["knn"] + "vector":
                    blobs.append(self._decode_descriptor(o))
                elif p == self.namespaces["knn"] + "distance":
                    command["distances"] = True
                    bindings.append((o, command, "_distance"))
                else:
                    s2 = knn_nodes[s]
                    command = commands[s2]
                    prop = self._parse_uri(p)[1]
                    if isinstance(o, Literal):
                        command[prop] = o.toPython()
                    else:
                        add_binding(s2, p, o)
            elif p == self.namespaces["c"] + "ANY":
                add_is_connected_to(s, None, o)
            else:
                raise SPARQLError(
                    f"Unknown property or connection: {p}: valid connections are {self.connections.keys()}, valid properties are {self.properties.keys()}")

        result, output, _ = execute_query(self._client, query, blobs)
        solutions = set()  # process_bindings uses this to avoid yielding duplicate solutions
        yield from process_bindings(output)

        if self.debug:  # Debugging
            self.bindings = bindings
            self.commands = commands
            self.input_query = query
            self.object_prefixes = object_prefixes
            self.output_response = output
            self.solutions = solutions
            self.triples = triples
        self.logger.info(
            f"Evaluated BGP query with {len(triples)} triples, produced {len(commands)} commands, and {len(solutions)} solutions")
        return

    def _decode_descriptor(self, literal: "Literal") -> bytes:
        import numpy as np
        return np.array(json.loads(literal.toPython())).astype(np.float32).tobytes()

    @classmethod
    def encode_descriptor(self, vector: "np.array") -> "Literal":
        from rdflib import Literal
        return Literal(json.dumps(vector.tolist()))

    def _optimize_query(self, query: List[dict], object_prefixes: List["URIRef"]) -> List[dict]:
        """
        Optimize the query by reordering Find commands

        Query should not have connections between Find commands yet
        """
        # Sort by number of constraints, descending, alphabetical by command name
        # TODO: Could use statistics from schema to optimize further
        query, object_prefixes = list(zip(*sorted(zip(query, object_prefixes),
                                                  key=lambda x: (-len(list(x[0].values())[0].get("constraints", [])), list(x[0].keys())[0]))))
        for i, command in enumerate(query, start=1):
            list(command.values())[0]["_ref"] = i
        return query, object_prefixes

    def _apply_context_to_triples(self, ctx: "QueryContext",
                                  triples: List[Tuple["Identifier",
                                                      "Identifier", "Identifier"]]
                                  ) -> List[Tuple["Identifier", "Identifier", "Identifier"]]:
        """
        Apply the query context to the triples
        """
        new_triples = []
        for s, p, o in triples:
            s2 = ctx[s]
            p2 = ctx[p]
            o2 = ctx[o]

            if s2 is not None:
                s = s2
            if p2 is not None:
                p = p2
            if o2 is not None:
                o = o2
            new_triples.append((s, p, o))
        return new_triples

    def _deduce_types(self, cts: "QueryContext",
                      triples: List[Tuple["Identifier",
                                          "Identifier", "Identifier"]]
                      ) -> Dict["Identifier", str]:
        """
        Deduce the type of each subject and connection object in the query
        """
        def add_types(k, tt):
            if k not in types:
                types[k] = tt
            else:
                intersection = types[k].intersection(tt)
                if not intersection:
                    raise SPARQLError(
                        f"Type error: {k} does not match a type {types[k]} vs {tt}")
                types[k] = intersection

        from rdflib import RDF
        from rdflib.plugins.sparql.sparql import QueryContext, SPARQLError
        types = {}
        for s, p, o in triples:
            s_type = self._deduce_type(s)
            if s_type:
                add_types(s, {s_type})

            if p in self.properties:
                add_types(s, self.properties[p])
            elif p in self.connections:
                src, dst = self.connections[p]
                o_type = self._deduce_type(o)
                if o_type:
                    add_types(o, [o_type])
                add_types(s, src)
                add_types(o, dst)
            elif p == RDF.type:
                add_types(s, [self._parse_uri_with_ns('t', o)])
            elif p in self.knn_properties:
                if p == self.namespaces["knn"] + "similarTo":
                    add_types(s, {"_Descriptor"})
            elif p == self.namespaces["c"] + "ANY":
                pass
            else:
                raise SPARQLError(
                    f"Unknown property or connection: {p}: valid connections are {self.connections.keys()}, valid properties are {self.properties.keys()}")

        for v, t in types.items():
            if len(t) > 1:
                raise SPARQLError(f"Ambiguous type for {v}: {t}")

        return {v: list(t)[0] for v, t in types.items()}

    def _deduce_type(self, v: "Identifier") -> str:
        """
        Deduce the type of an identifier based on its URI
        """
        from rdflib.term import URIRef
        if isinstance(v, URIRef) and v.startswith(self.namespaces["o"]):
            local_name = self._parse_uri_with_ns("o", v)
            return local_name.split("/")[0]
        return None

    def _deduce_uniqueid(self, v: "Identifier") -> str:
        """
        Deduce the uniqueid of an identifier based on its URI
        """
        from rdflib.term import URIRef
        if isinstance(v, URIRef) and v.startswith(self.namespaces["o"]):
            local_name = self._parse_uri_with_ns("o", v)
            return local_name.split("/")[1]
        return None

    def query(self, query: str):
        """
        Execute a SPARQL query
        """
        return self.graph.query(query)

    def to_dataframe(self, result) -> "pd.DataFrame":
        """
        Convert the SPARQL result to a pandas DataFrame
        """
        import pandas as pd

        return pd.DataFrame(
            data=([None if x is None else self._format_node(x)
                   for x in row] for row in result),
            columns=[str(x) for x in result.vars],
        )

    def get_blob(self, uri: Union[str, "URIRef"], type: Optional[str] = None) -> bytes:
        """
        Get the blob associated with a URI or QName.
        """
        if isinstance(uri, str):
            uri = self.graph.namespace_manager.expand_curie(uri)

        if type is None:
            type = self._deduce_type(uri)
        else:
            assert type == self._deduce_type(
                uri), f"Type {type} does not match deduced type {self._deduce_type(uri)}"
        assert type is not None, f"Cannot get blob for entity URI: {uri}"
        assert type[0] == "_", f"Cannot get blob for entity URI: {uri} with type {type}"
        uniqueid = self._deduce_uniqueid(uri)
        command_name = "Find" + type[1:]
        query = [
            {command_name: {
                "constraints": {
                    "_uniqueid": ["==", uniqueid],
                },
                "blobs": True,
            }},
        ]
        _, blobs = self._client.query(query)
        return blobs[0]

    def get_blobs(self, uris: List[Union[str, "URIRef"]], type: Optional[str] = None) -> List[bytes]:
        """
        Get the blobs associated with a list of URI or QName.
        """
        uris = [self.graph.namespace_manager.expand_curie(uri)
                if isinstance(uri, str) else uri for uri in uris]
        types = [self._deduce_type(uri) for uri in uris]
        if type is not None:
            assert all(
                t == type for t in types), f"Type {type} does not match deduced types {types}"
        else:
            type = types[0]
            assert all(
                t == type for t in types), f"Types do not match: {types}"

        assert type is not None, f"Cannot get blob for entity URI: {uri}"
        assert type[0] == "_", f"Cannot get blob for entity URI: {uri} with type {type}"
        uniqueids = [self._deduce_uniqueid(uri) for uri in uris]
        command_name = "Find" + type[1:]
        query = [
            {command_name: {
                "constraints": {
                    "_uniqueid": ["in", uniqueids],
                },
                "blobs": True,
            }},
        ]
        _, blobs = self._client.query(query)
        return blobs

    def get_image(self, uri: Union[str, "URIRef"]) -> "np.ndarray":
        """
        Get the image associated with a URI or QName
        """
        import numpy as np
        import cv2

        blob = self.get_blob(uri, type="_Image")
        nparr = np.fromstring(blob, np.uint8)
        image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        return image

    def get_images(self, uris: List[Union[str, "URIRef"]]) -> List["np.ndarray"]:
        """
        Get the images associated with a list of URI or QName
        """
        import numpy as np
        import cv2

        blobs = self.get_blobs(uris, type="_Image")
        images = []
        for blob in blobs:
            nparr = np.fromstring(blob, np.uint8)
            image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            images.append(image)
        return images

    def show_images(self, uris: List[Union[str, "URIRef"]]):
        """
        Show the images associated with a list of URI or QName

        This is best used in a Jupyter notebook.
        """
        import matplotlib.pyplot as plt

        images = self.get_images(uris)
        columns = min(5, len(images))
        rows = math.ceil(len(images) / columns)
        for i, image in enumerate(images):
            plt.subplot(rows, columns, i + 1)
            plt.imshow(image)
            plt.axis("off")
        plt.show()

    def get_descriptor(self, uri: Union[str, "URIRef"]) -> "np.ndarray":
        """
        Get the descriptor associated with a URI or QName
        """
        import numpy as np
        blob = self.get_blob(uri, type="_Descriptor")
        return np.frombuffer(blob, dtype=np.float32)

    def get_descriptors(self, uris: List[Union[str, "URIRef"]]) -> List["np.ndarray"]:
        """
        Get the descriptors associated with a list of URI or QName
        """
        import numpy as np
        blobs = self.get_blobs(uris, type="_Descriptor")
        return [np.frombuffer(blob, dtype=np.float32) for blob in blobs]
